import os


settings = {
    'logging': {
        'version': 1,
        'disable_existing_loggers': True,
        'formatters': {
            'standard': {
                'format': '%(asctime)s [%(levelname)s] %(message)s',
                'datefmt': '%H:%M:%S'
            }
        },
        'handlers': {
            'console': {
                'level': 'DEBUG',
                'class': 'logging.StreamHandler',
                'formatter': 'standard',
            },
        },
        'loggers': {
            '': {
                'handlers': ['console'],
                'level': 'DEBUG',
                'propagate': True
            },
        }
    },
    'pool_server': {
        'ip': '10.5.0.2',
        'port': 8000
    },
    'log_server': {
        'ip': '10.5.0.3',
        'port': 8000
    },
    'miner_server': {
        'ip': '10.5.0.4',
        'port': 8000
    }
}

if os.getenv('IS_LOCAL', False):
    settings['pool_server'] = {'ip': '127.0.0.1', 'port': 8000}
    settings['log_server'] = {'ip': '127.0.0.1', 'port': 8001}
    settings['miner_server'] = {'ip': '127.0.0.1', 'port': 8002}



import logging
import random
import requests

import tornado.ioloop
import tornado.web

from requests.exceptions import ConnectionError

from examples.common.conf import settings
from examples.common.utils import init_logger

from openchain.models.wallet import Wallet
from openchain.models.transaction import Transaction

logger = logging.getLogger()


def generate_and_send_transaction(wallet, miner_list):
    transaction = Transaction(in_address=wallet.address, out_address='', amount=random.randint(0, 50))
    transaction.signing(wallet.private_key_hex)
    if transaction.is_valid:
        for miner in miner_list:
            r = requests.post(miner['client_id'], json=transaction.__dict__)
            if r.status_code != 200:
                logger.debug('[WALLET] Error sending transaction, server not responding')
                return

            result = r.json()
            if result['status'] == 200:
                logger.debug('[WALLET] Successfully send transaction to {}'.format(miner['client_id']))
            else:
                logger.debug('[WALLET] Error sending transaction: {}'.format(result['message']))
            return


if __name__ == "__main__":
    init_logger(settings)
    logger.debug('[WALLET] Starting wallet application')

    wallet_list = Wallet.objects.get()
    if not len(wallet_list):
        logger.debug('[WALLET] Creating new wallet')
        wallet = Wallet()
        wallet.save()
    else:
        logger.debug('[WALLET] Using existing wallet')
        wallet = wallet_list[0]

    logger.debug('[WALLET] Connecting to pool server')
    shutdown = False
    try:
        r = requests.get('http://{}:{}'.format(settings['pool_server']['ip'],
                                               settings['pool_server']['port']))
        if r.status_code != 200:
            logger.error('[MINER] Pool server is currently unavailable')
            shutdown = True
        else:
            result = r.json()
            if result['status'] != 200:
                logger.error('[WALLET] Pool server encountered error {}'.format(result['message']))
                shutdown = True
            elif not result['data']:
                logger.error('[WALLET] There are no available miner servers found')
                shutdown = True
    except ConnectionError:
        logger.error('[WALLET] Pool server is currently unavailable')
        shutdown = True

    if shutdown:
        logger.debug('[WALLET] Shutdown the application')
        exit(-1)

    main_loop = tornado.ioloop.IOLoop.instance()
    scheduled_loop = tornado.ioloop.PeriodicCallback(lambda: generate_and_send_transaction(wallet, result['data']),
                                                     7000, io_loop=main_loop)

    logger.debug('[WALLET] Start transaction generation')
    scheduled_loop.start()
    main_loop.start()



import logging
import requests

import tornado.ioloop
import tornado.web

from requests.exceptions import ConnectionError

from examples.common.conf import settings
from examples.common.utils import init_logger
from examples.listener.miner import MinerListener

from openchain.models.block import Block
from openchain.models.transaction import Transaction

logger = logging.getLogger()


def generate_blockchain():
    transaction_list = []
    for transaction in Transaction.objects.get():
        transaction_list.append(transaction.__dict__)

    if len(transaction_list):
        block = Block(Block.objects.blockchain.last_block_hash, transactions=transaction_list)
        block.generate()
        block.save()

        logger.debug('[MINER] Added block with {} transactions'.format(len(transaction_list)))
        Transaction.objects.delete_all()
    else:
        logger.debug('[MINER] Skipping empty block generation')


if __name__ == "__main__":
    init_logger(settings)
    logger.debug('[MINER] Starting the application')

    blockchain = Block.objects.blockchain
    logger.debug('[MINER] Blockchain loaded with {} blocks'.format(len(blockchain.block_list)))

    logger.debug('[MINER] Announcing address to pool server')
    headers = {'X-Client-Address': 'http://{}:{}'.format(settings['miner_server']['ip'],
                                                         settings['miner_server']['port'])}
    shutdown = False
    try:
        r = requests.post('http://{}:{}'.format(settings['pool_server']['ip'],
                                                settings['pool_server']['port']), headers=headers)
        if r.status_code != 200:
            logger.error('[MINER] Pool server is currently unavailable')
            shutdown = True
        else:
            result = r.json()
            if result['status'] != 200:
                logger.error('[MINER] Pool server encountered error {}'.format(result['message']))
                shutdown = True
    except ConnectionError:
        logger.error('[MINER] Pool server is currently unavailable')
        shutdown = True

    if shutdown:
        logger.debug('[MINER] Shutdown the application')
        exit(-1)

    app = tornado.web.Application([
        (r"/", MinerListener),
    ])
    app.listen(settings['miner_server']['port'], address=settings['miner_server']['ip'])

    main_loop = tornado.ioloop.IOLoop.instance()
    scheduled_loop = tornado.ioloop.PeriodicCallback(generate_blockchain, 60000, io_loop=main_loop)

    logger.debug('[MINER] Start blockchain generation')
    scheduled_loop.start()

    logger.debug('[MINER] Listening for connections on {}:{}'.format(settings['miner_server']['ip'],
                                                                     settings['miner_server']['port']))
    main_loop.start()



import os
import leveldb


class Singleton(type):

    instance_set = {}

    def __call__(cls, *args, **kwargs):
        if cls not in cls.instance_set:
            cls.instance_set[cls] = super(Singleton, cls).__call__(*args, **kwargs)
        return cls.instance_set[cls]


class LevelDBAdapter(metaclass=Singleton):

    cs = {}

    def connect(self, namespace):
        db_path = os.getenv('DATABASE_PATH', '/var/tmp/leveldb/')
        if namespace not in self.cs:
            filename = os.path.join(db_path, namespace)
            self.cs[namespace] = leveldb.LevelDB(filename)
        return self.cs[namespace]

    @staticmethod
    def write_batch():
        return leveldb.WriteBatch()



import logging

import tornado.ioloop
import tornado.web

from examples.common.conf import settings
from examples.common.utils import init_logger
from examples.listener.pool import PoolListener

logger = logging.getLogger()


if __name__ == "__main__":
    init_logger(settings)
    logger.debug('[POOL] Starting pool application')

    app = tornado.web.Application([
        (r"/", PoolListener),
    ])

    app.listen(settings['pool_server']['port'], address=settings['pool_server']['ip'])

    logger.debug('[POOL] Listening for connections on {}:{}'.format(settings['pool_server']['ip'],
                                                                    settings['pool_server']['port']))
    tornado.ioloop.IOLoop.current().start()



class ModelFactory:

    @staticmethod
    def get_model(name):
        from openchain.models.block import Block
        from openchain.models.client import Client
        from openchain.models.logentry import LogEntry
        from openchain.models.transaction import Transaction
        from openchain.models.wallet import Wallet

        model_map = {
            'block': Block,
            'client': Client,
            'logentry': LogEntry,
            'transaction': Transaction,
            'wallet': Wallet,
        }

        if name in model_map:
            return model_map[name]
        return None



import json
import logging
import tornado.web

from openchain.models.client import Client

logger = logging.getLogger()


class PoolListener(tornado.web.RequestHandler):

    def get(self):
        logger.debug('[POOL] Processing get request')

        try:
            result = {
                'status': 200,
                'data': Client.objects.dict_list
            }
        except Exception as e:
            logger.error('[POOL] {}'.format(e))
            result = {'status': 500, 'message': e}

        self.set_header('Content-Type', 'application/json')
        self.write(json.dumps(result).encode())
        self.finish()

    def post(self):
        logger.debug('[POOL] Processing post request')

        try:
            client_id = self.request.headers.get('X-Client-Address', self.request.remote_ip)
            client = Client(client_id)
            client.save()
            result = {'status': 200, 'message': 'OK'}
        except Exception as e:
            logger.error('[POOL] {}'.format(e))
            result = {'status': 500, 'message': e}

        self.set_header('Content-Type', 'application/json')
        self.write(json.dumps(result).encode())
        self.finish()



import json
import logging
import tornado.web

from openchain.models.block import Block
from openchain.models.transaction import Transaction

logger = logging.getLogger()


class MinerListener(tornado.web.RequestHandler):

    def get(self):
        logger.debug('[MINER] Processing get request')

        try:
            blockchain = Block.objects.blockchain
            result = {
                'status': 200,
                'is_valid': blockchain.is_valid,
                'blockchain': blockchain.__dict__
            }
        except Exception as e:
            logger.error('[MINER] {}'.format(e))
            result = {'status': 500, 'message': e}

        self.set_header('Content-Type', 'application/json')
        self.write(json.dumps(result).encode())
        self.finish()

    def post(self):
        logger.debug('[MINER] Processing post request')

        try:
            result = {'status': 204, 'message': 'No Content'}
            if self.request.body:
                transaction = json.loads(self.request.body.decode('utf-8'))
                transaction = Transaction(**transaction)
                transaction.save()
                result = {'status': 200, 'message': 'OK'}
        except Exception as e:
            logger.error('[MINER] {}'.format(e))
            result = {'status': 500, 'message': e}

        self.set_header('Content-Type', 'application/json')
        self.write(json.dumps(result).encode())
        self.finish()



import logging

from socket import socket

from openchain.utils.stun import get_ip_info, OpenInternet, FullCone

logger = logging.getLogger()


def init_socket():
    nat_type, ip, port = get_ip_info()
    if nat_type not in (OpenInternet, FullCone):
        logger.critical('Your internet connection does not support NAT translation')
        logger.info('Trying to use internal routing')

    if ip is None or port is None:
        logger.critical('Cant get external client address')
        local = True

    if local:
        ip = '0.0.0.0'
        port = 8113

    logger.info('Trying to bind address {}:{}'.format(ip, port))

    sock = socket()
    sock.bind((ip, port))

    return sock, ip, port


def get_client_id(request)->str:
    return request.headers.get('X-Client-STUN-Address', '{}:8112'.format(request.remote_ip))



import os

role = os.getenv('ROLE')
if role in ['pool', 'miner', 'wallet', 'logger']:
    os.system('python3 {}.py'.format(role))
else:
    print('OS env var ROLE -{}- not set'.format(role))
    exit(-1)



import json
import logging
import tornado.web

from openchain.models.logentry import LogEntry

logger = logging.getLogger()


class LoggerListener(tornado.web.RequestHandler):

    def get(self):
        logger.debug('[LOGGER] Processing get request')

        try:
            result = {
                'status': 200,
                'data': LogEntry.objects.dict_list
            }
        except Exception as e:
            logger.error('[LOGGER] {}'.format(e))
            result = {'status': 500, 'message': e}

        self.set_header('Content-Type', 'application/json')
        self.write(json.dumps(result).encode())
        self.finish()

    def post(self):
        logger.debug('[LOGGER] Processing post request')

        try:
            result = {'status': 204, 'message': 'No Content'}
            if self.request.body:
                log_entry = LogEntry(str(self.request.body.decode('utf-8')))
                log_entry.save()
                result = {'status': 200, 'message': 'OK'}
        except Exception as e:
            logger.error('[LOGGER] {}'.format(e))
            result = {'status': 500, 'message': e}

        self.set_header('Content-Type', 'application/json')
        self.write(json.dumps(result).encode())
        self.finish()



import collections

from openchain.models.exception import BlockchainTreeChildCollisionException, BlockchainTreeParentCollisionException


class BlockchainNode:

    block = None
    prev_item = None
    next_item = None
    depth = 0

    def __init__(self, block: callable, prev_item: callable=None, next_item: callable=None):
        self.block = block
        self.prev_item = prev_item
        self.next_item = next_item
        self.depth = 0

    @property
    def __dict__(self):
        unordered = {
            'block': self.block.__dict__,
            'prev_item': self.prev_item.block.data_hash if self.prev_item else None,
            'next_item': self.next_item.block.data_hash if self.next_item else None,
        }
        return collections.OrderedDict(sorted(unordered.items()))

    def calculate_depth(self) -> int:
        if not self.next_item:
            return 0
        return self.next_item.calculate_depth() + 1


class Blockchain:

    collisions = []
    block_tree = {}
    block_list = []

    def __init__(self, block_list: list):
        self.collisions = []
        self.block_tree = {}
        self.block_list = block_list

    @property
    def is_valid(self) -> bool:
        return len(self.collisions) == 0

    @property
    def __dict__(self):
        result = {}
        for block_hash, block in self.block_tree.items():
            result[block_hash] = block.__dict__
        return result

    @property
    def last_block_hash(self):
        try:
            block_hash = next(iter(self.block_tree))
            return self.get_latest_block_hash(block_hash)
        except StopIteration:
            return ''

    def get_latest_block_hash(self, hash):
        if self.block_tree[hash].next_item is None:
            return hash
        else:
            next_hash = self.block_tree[hash].next_item.block.data_hash
            return self.get_latest_block_hash(next_hash)

    def generate_tree(self, raise_exception: bool=True):
        for block in self.block_list:
            curr_block_node = BlockchainNode(block)

            if block.prev_block in self.block_tree:
                prev_block_node = self.block_tree[block.prev_block]
                if prev_block_node.next_item:
                    self.collisions.extend([block, prev_block_node.block])
                    if raise_exception:
                        raise BlockchainTreeChildCollisionException
                    continue

                prev_block_node.next_item = curr_block_node
                curr_block_node.prev_item = prev_block_node

            if block.next_block in self.block_tree:
                next_block_node = self.block_tree[block.next_block]
                if next_block_node.prev_item:
                    self.collisions.extend([block, next_block_node.block])
                    if raise_exception:
                        raise BlockchainTreeParentCollisionException
                    continue

                next_block_node.prev_item = curr_block_node
                curr_block_node.next_item = next_block_node

            self.block_tree[block.data_hash] = curr_block_node

        for key, node in self.block_tree.items():
            node.depth = node.calculate_depth()



import collections
import time

from openchain.models.base import Model, Manager


class LogEntryManager(Manager):
    pass


class LogEntry(Model):

    message = None
    timestamp = None

    objects = LogEntryManager()

    def __init__(self, message: str, timestamp: float=None):
        self.message = message
        if timestamp is None:
            self.timestamp = time.time()

    @property
    def __dict__(self):
        unordered = {
            'message': self.message,
            'timestamp': self.timestamp
        }
        return collections.OrderedDict(sorted(unordered.items()))



import logging
import logging.config

from array import array

logger = logging.getLogger()


def string_to_bytes(string):
    array_key = array('b')
    array_key.frombytes(string.encode())
    return array_key.tobytes()


def init_logger(settings):
    logging.basicConfig(level=logging.DEBUG)
    logging.config.dictConfig(settings['logging'])
    return logger


def get_address():
    ip = port = interface = None
    return ip, port, interface



class TransactionInvalidPublicKeyException(Exception):
    """Transaction has invalid public key"""


class TransactionInvalidSignatureException(Exception):
    """Transaction has invalid signature"""
